#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Source code for tested Ordinary least--squares solvers.
Of the methods, __rankKrandomSVD was the only one without a ready implementation.

Submitted 6.11.2020 to Machine Learning and Knowledge Extraction as
supplementary material for manuscript:
    Do Randomized Algorithms Improve the Efficiency of Minimal Learning Machine?

@author: Joakim Linja

Package versions:

numpy==1.16.4
scikit-learn==0.20.3
scipy==1.3.0
"""

from numpy.random import normal
from numpy import (matrix as npmat, multiply,
                   eye, sqrt, finfo, array)
from numpy.linalg import (qr, svd as npsvd, inv,
                          solve, cholesky,
                          lstsq as nplstsq)
from scipy.linalg import (solve_triangular as soltri, svd as spsvd)
from sklearn.utils.extmath import randomized_svd
from functools import partial


def __rankKrandomSVD(A: npmat, k: int):
    """
    Implemented using source material.
    https://amath.colorado.edu/faculty/martinss/Talks/2010_mmds_martinsson.pdf
    """
    _, n = A.shape
    # 1. O: n x k
    O = normal(size=(n,k))
    # 2. Y = A @ O, Y: m x k
    #Y = A @ O
    # 3. Q: m x k, Y = Q @ R
    Q, _ = qr(A @ O, mode='reduced')
    # 4. B = Q @ A, B: k x n
    #B = Q @ A
    # 5. B = U1 @ S @ Vh
    U1, S, Vh = npsvd(Q.T @ A, full_matrices=False)
    # 6. U = Q @ U1
    return Q @ U1, S, Vh


def np_inverse_NR(Dx: npmat, Dy: npmat):
    """
    B = ( Dx.T @ Dx)^-1 @ Dx.T @ Dy
    """
    return inv(Dx.T @ Dx) @ Dx.T @ Dy


def np_inverse(Dx: npmat, Dy: npmat):
    """
    B = ( Dx.T @ Dx)^-1 @ Dx.T @ Dy
    """
    dig = sqrt(finfo(float).eps)*eye(Dx.shape[1])
    return inv(Dx.T @ Dx + dig) @ Dx.T @ Dy


def np_solve_NR(Dx: npmat, Dy: npmat):
    """
    (Dx.T @ Dx) @ B = Dx.T @ Dy
    """
    return solve(Dx.T @ Dx, Dx.T @ Dy)


def np_solve(Dx: npmat, Dy: npmat):
    """
    (Dx.T @ Dx) @ B = Dx.T @ Dy
    """
    dig = sqrt(finfo(float).eps)*eye(Dx.shape[1])
    return solve(Dx.T @ Dx + dig, Dx.T @ Dy)


def np_svd(Dx: npmat, Dy: npmat):
    """
    (U, S, V.T) = Dx
    B = V @ S^-1 @ U @ Dy
    """
    (U,S,Vh) = npsvd(Dx, full_matrices=False)
    return multiply(Vh.T, 1/S) @ U.T @ Dy


def sp_svd(Dx: npmat, Dy: npmat):
    """
    (U, S, V.T) = Dx
    B = V @ S^-1 @ U @ Dy
    """
    (U,S,Vh) = spsvd(Dx, full_matrices=False,compute_uv=True,
        overwrite_a=False,check_finite=False)
    return multiply(Vh.T, 1/S) @ U.T @ Dy


def rankKrandomSVD(p: float, Dx: npmat, Dy: npmat):
    """
    (U, S, V.T) ~= Dx
    B = V @ S^-1 @ U @ Dy
    """
    (U,S,Vh) = __rankKrandomSVD(Dx, int(Dx.shape[0]*p))
    return multiply(Vh.T, 1/S) @ U.T @ Dy


def sk_rand_svd(p: float, I, Dx, Dy):
    """
    (U, S, V.T) ~= Dx
    B = V @ S^-1 @ U @ Dy
    """
    U,S,Vh = randomized_svd(Dx,
                n_components=(int(Dx.shape[0]*p)),n_iter=I)
    return multiply(Vh.T, 1/S) @ U.T @ Dy


def cholesky_decomposition(Dx: npmat, Dy: npmat, pieceR=False):
    """
    L @ L* = Dx.T @ Dx
    L @ y = I
    L* @ (Dx.T @ Dx)^-1 = y



    Dx = L @ L*
    Dx.T = (L*).T @ L.T
    Dx.T @ Dx = (L*).T @ L.T @ L @ L*
              = the same
    """
    if type(Dx)!=npmat:
        Dx = npmat(Dx)
    L = cholesky(Dx.T @ Dx)
    imat = soltri(L.H, soltri(L, eye(L.shape[1]),lower=True))
    if pieceR:
        return imat, Dx.T, Dy
    else:
        return imat @ Dx.T @ Dy


def lstsq(Dx: npmat, Dy: npmat):
    """
    (Dx.T @ Dx + ~eps*I) @ B = Dx.T @ Dy
    """
    Dx = array(Dx)
    Dy = array(Dy)
    dig = sqrt(finfo(float).eps)*eye(Dx.shape[1])
    return nplstsq(Dx.T @ Dx + dig, Dx.T @ Dy, rcond=None)[0]


def lstsq_NR(Dx: npmat, Dy: npmat):
    """
    (Dx.T @ Dx) @ B = Dx.T @ Dy
    """
    return nplstsq(Dx.T @ Dx, Dx.T @ Dy, rcond=None)[0]


def getBcalcMethodDict():
    cmd = {'np_inverse_NR':np_inverse_NR,
           'np_inverse':np_inverse,
           'np_solve_NR':np_solve_NR,
           'np_solve':np_solve,
           'np_svd':np_svd,
           'sp_svd':sp_svd,
           'cholesky_decomposition':cholesky_decomposition,
           'rankKrandomSVD_02':partial(rankKrandomSVD,0.2),
           'rankKrandomSVD_04':partial(rankKrandomSVD,0.4),
           'rankKrandomSVD_06':partial(rankKrandomSVD,0.6),
           'rankKrandomSVD_08':partial(rankKrandomSVD,0.8),
           'rankKrandomSVD_10':partial(rankKrandomSVD,1.0),
           'sk_rank_svd_02':partial(sk_rand_svd,0.2,1),
           'sk_rank_svd_04':partial(sk_rand_svd,0.4,1),
           'sk_rank_svd_06':partial(sk_rand_svd,0.6,1),
           'sk_rank_svd_08':partial(sk_rand_svd,0.8,1),
           'sk_rank_svd_10':partial(sk_rand_svd,1.0,1),
           'lstsq':lstsq,
           'lstsq_NR':lstsq_NR,
           'sk_rank_svd_001':partial(sk_rand_svd,0.01,1),
           'sk_rank_svd_002':partial(sk_rand_svd,0.02,1),
           'sk_rank_svd_003':partial(sk_rand_svd,0.03,1),
           'sk_rank_svd_004':partial(sk_rand_svd,0.04,1),
           'sk_rank_svd_005':partial(sk_rand_svd,0.05,1),
           'sk_rank_svd_006':partial(sk_rand_svd,0.06,1),
           'sk_rank_svd_007':partial(sk_rand_svd,0.07,1),
           'sk_rank_svd_008':partial(sk_rand_svd,0.08,1),
           'sk_rank_svd_009':partial(sk_rand_svd,0.09,1),
           'sk_rank_svd_010':partial(sk_rand_svd,0.10,1),
           }
    return cmd
