#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Source code for Minimal Learning Machine and RS-maximin.

Original implementation by Joonas Hämäläinen
Modified by Joakim Linja:
    train-function now returns Dx and Dy instead of B so that
    different B-calculation methods can be tested
    pred-function input argument order has been changed to allow partial function use

Submitted 6.11.2020 to Machine Learning and Knowledge Extraction as
supplementary material for manuscript:
    Do Randomized Algorithms Improve the Efficiency of Minimal Learning Machine?

@author: Joonas Hämäläinen
@author: Joakim Linja
"""

#
# Minimal Learning Machine (MLM) implementation for 1D regression
#

import numpy as np
from sklearn.metrics.pairwise import euclidean_distances
import warnings
import random

# Train MLM
def train(X,Y,K_rel=100,rs_method="random", printP=True,
          returnRefpIndices=False, forceRS=False):
    """
    Trains MLM model.

    Parameters
    ----------

    X: numpy.ndarray
        Input data matrix, rows are observations.

    Y: numpy.ndarray
        Output data matrix, rows are observations.

    K_rel: int, optional
        Relative #observations to be seleted as reference points.

        Default: 100

    rs_method: str, optional
        "random" uses random sampling to select reference points.

        "RS-maximin" uses RS-maximin to select reference points.

        Default: "random"


    Returns
    -------

    Dx: numpy.matrix
        Input space distance matrix

    Dy: numpy.matrix
        Output space distance matrix

    R: numpy.ndarray
        Input data reference points

    T: numpy.ndarray
        Output data reference points

    """

    # X and Y should be 2D ndarrays, e.g., shape should be (N,1)
    # not (N,)
    if len(X.shape) == 1:
        warnings.warn("Reshaping input data matrix...")
        X = X.reshape(-1,1)
    if len(Y.shape) == 1:
        warnings.warn("Reshaping output data matrix...")
        Y = Y.reshape(-1,1)

    # Check for constant variables
    X_max = np.max(X,axis=0)
    X_min = np.min(X,axis=0)
    is_const_var = X_max - X_min == 0

    if type(is_const_var)==np.ndarray:
        if any(is_const_var):
            warnings.warn("Input data matrix contains constant columns!" \
                      "Results may be inaccurate!")
    else:
        if any(list(map(lambda x: any(x), np.array(is_const_var)))):
            warnings.warn("Input data matrix contains constant columns!" \
                      "Results may be inaccurate!")

    if printP:
        print("Selecting reference points with " + rs_method + " method...")

    if not returnRefpIndices:
        R,T = rps(X,Y,K_rel,rs_method,forceRS=forceRS)

        Dx, Dy, K = _train(X,Y,R,T)

        return Dx, Dy, R, T
    else:
        R, T, I = rps(X,Y,K_rel,rs_method,
                returnRefpIndices=returnRefpIndices,
                forceRS=forceRS)

        return X, Y, R, T, I


def train_midphase(X, Y, R, T, I):
    """
    """
    r = R[I[:,0]]
    t = T[I[:,0]]
    Dx, Dy, _ = _train(X, Y, r, t)
    return Dx, Dy, r, t


# Trains MLM with selected sets of reference points
def _train(X,Y,R,T):
    """
    Trains MLM model.

    Parameters
    ----------

    X: numpy.ndarray
        Input data matrix, rows are observations.

    Y: numpy.ndarray
        Output data matrix, rows are observations.

    R: numpy.ndarray
        Input data reference points

    T: numpy.ndarray
        Output data reference points


    Returns
    -------

    B: numpy.ndarray
        Distance regression model coefficients.

    """

    # X,Y,R,T should be 2D ndarrays also for 1D cases,
    # e.g., shape should be (N,1) not (N,)
    if len(X.shape) == 1:
        warnings.warn("Reshaping input data matrix...")
        X = X.reshape(-1,1)
    if len(Y.shape) == 1:
        warnings.warn("Reshaping output data matrix...")
        Y = Y.reshape(-1,1)
    if len(R.shape) == 1:
        warnings.warn("Reshaping input data reference point matrix...")
        R = R.reshape(-1,1)
    if len(T.shape) == 1:
        warnings.warn("Reshaping output data reference point matrix...")
        T = T.reshape(-1,1)

    # Compute the distance matrices
    Dx = euclidean_distances(X,R)
    Dy = euclidean_distances(Y,T)

    # Compute B = (Dx^T Dx)^-1 (Dx^T Dy)
    K = len(T)

    # Original:
    """
    DxDx = np.matmul(Dx.transpose(),Dx) + np.sqrt(np.finfo(float).eps)*np.eye(K)
    DxDy = np.matmul(Dx.transpose(),Dy)

    B = np.linalg.lstsq(np.matmul(Dx.transpose(),Dx) +\
                        np.sqrt(np.finfo(float).eps)*np.eye(K),\
                        np.matmul(Dx.transpose(),Dy))[0]
    return B
    """
    return Dx, Dy, K


# Reference Points Selection for MLM
def rps(X, Y, K_rel, rs_method, returnRefpIndices=False,
        forceRS=False):
    """
    Selects reference points for the MLM model.

    Parameters
    ----------

    X: numpy.ndarray
        Input data matrix, rows are observations.

    Y: numpy.ndarray
        Output data matrix, rows are observations.

    K_rel: int
        Relative #observations to be seleted as reference points.

    rs_method: str
        "random" uses random sampling to select reference points. \
        "RS-maximin" uses RS-maximin to select reference points.


    Returns
    -------

    R: numpy.ndarray
        Input data reference points

    T: numpy.ndarray
        Output data reference points

    """
    N = len(Y)
    K_abs = int(np.round(N*K_rel/100))

    if K_abs == N and (not forceRS): # Full MLM
        R = X
        T = Y
    elif rs_method == "random":
        I = np.random.permutation(N)
        R = X[I[range(K_abs)]]
        T = Y[I[range(K_abs)]]
    elif rs_method == "RS-maximin":
        I = np.zeros((K_abs,1),dtype=int)

        # First ref. point closest to data mean
        xm = np.mean(X,axis=0).reshape(1,-1)
        I[0] = np.argmin(euclidean_distances(xm,X))

        i = 1
        fill_rest = False
        D = np.full(N,float("inf"))
        while i < K_abs:
            D2 = euclidean_distances(X,X[I[i-1]]).reshape(1,-1)
            D = np.min(np.vstack((D,D2)),axis=0)
            Id = np.where(D > 0)
            Id = Id[0]
            maxi = np.argmax(D[Id])
            if len(Id) > 0:
                I[i] = Id[maxi]
                i += 1
            else:
                fill_rest = True
                break

        # Add duplicate reference points
        if fill_rest:
            I_dup = np.setdiff1d(np.array(range(N)).reshape(-1,1),I)
            I[i:K_abs,:] = I_dup.reshape(-1,1)

        R = X[I[:,0]]
        T = Y[I[:,0]]
    else:
        raise NotImplementedError

    if not returnRefpIndices:
        return R,T
    else:
        return R, T, I

# Predict with MLM
def pred(B,R,T,x,cubic=True):
    """
    Computes MLM prediction y for given input x.


    Parameters
    ----------
    x: numpy.ndarray
        Input vector

    B_hat: numpy.ndarray
        Distance regression model coefficients.

    R: numpy.ndarray
        Input data reference points

    T: numpy.ndarray
        Output data reference points

    Returns
    -------
    y_hat: numpy.ndarray
        Predicted output.
    """

    # Predict euclidean distances to T in output space
    delta_hat = np.matmul(euclidean_distances(x,R),B).reshape(-1,1)

    # Solve multilateration problem
    if cubic:
        y_hat = cubic_eq_solver(T, delta_hat)
    else:
        y_hat = LLS(T, delta_hat)

    return np.real(y_hat)

# Analytical Multilateration Solver for the MLM
def cubic_eq_solver(T, delta_hat):
    """
    Solves multilateration problem as a cubic equation using np.roots
    function. See equation (11) and algorithm 2 in Mesquita et al. "Ensemble
    of Efficient Minimal Learning Machines for Classification and Regression"
    2017.

    Parameters
    ----------
    T: numpy.ndarray
        Output data reference points

    delta_hat: numpy.ndarray
        Predicted euclidean distances to reference points in output space.

    Returns
    -------
    y_hat: numpy.ndarray
        Root corresponding to smallest cost function value.
    """
    delta_hat2 = delta_hat**2

    # Third order polynomial coefficients for the deritative of the cost
    # function (11)
    A = np.size(delta_hat)
    B = -3*np.sum(T)
    C = np.sum(3*(T**2)-delta_hat2)
    D = np.sum(delta_hat2*T-T**3)

    # Find three roots
    y = np.roots([A,B,C,D])
    I = np.array([np.isreal(y[0]), np.isreal(y[1]), np.isreal(y[2])])

    # Select a root corresponding to smallest cost function value
    Jy = np.full(3,float("inf"))
    if y[0] != y[1] and I[0] and I[1]:
        for j in range(3):
            Jy[j] = np.sum(((y[j]-T)**2-delta_hat2)**2)
        ind = np.argmin(Jy)
        y_hat = y[ind]
    else:
        y_hat = y[I]
        y_hat = y_hat[0]
    return y_hat


def LLS(T, delta_hat):
    """
    """
    (K,S) = T.shape
    a = np.arange(K)
    random.shuffle(a)
    ind = a[0]
    r = T[ind]
    dr = delta_hat[ind][0]

    ta = np.arange(K)
    T = T[ta!=ind]
    dha = np.arange(len(delta_hat))
    delta_hat = delta_hat[dha!=ind]
    K -=1

    A = T-r

    d_ir = euclidean_distances(T.reshape(-1,1),r.reshape(-1,1))

    b = 0.5*(np.power(dr.reshape(-1,1),2)+np.power(d_ir,2)-np.power(delta_hat.reshape(-1,1),2))

    theta = np.linalg.lstsq(A.reshape(-1,1), b, rcond=None)[0]
    return (theta.T + r)[0][0]
